;****************************************************************
;*  GEBUEHR.ASM							*
;*			Gebhrenzhler ohne Pulse		*
;*			=========================		*
;*								*
;*  05.10.98	Bodo Wenzel	V1.0, Erstellung		*
;****************************************************************

	title		"Gebhrenzhler"

VERSION	equ	H'10'

;================================================================
;	Auswahl:			"0#### #.##"
;	L	Eintrag blttern
;	M	=> Editieren
;	R	=> Zhlen
;	Editieren:			"0#### #.##" + Cursor
;	L	Ziffer "blttern" = zyklisch von 0..9	
;	M	=> Lschen/Abbruch/Neu
;	R	Nchste Stelle whlen, nach letzter => Auswahl
;	Lschen/Abbruch/Neu:		"L ABR NEU"
;	L	Lschen, dann => Auswahl
;	M	Abbruch, dann => Auswahl
;	R	Neuer Eintrag, dann => Editieren
;	Zhlen:				"#:## ##.##" + Blinken
;	L	=> Auswahl
;	M	(unbenutzt)
;	R	=> Warten
;	Warten:				"#:## ##.##"
;	L	=> Auswahl
;	M	(unbenutzt)
;	R	=> Zhlen
;
;	Bemerkungen:
;	- der erste Eintrag kann nicht gelscht werden.
;	- wenn der Speicher voll ist (21 Eintrge), wird kein
;	  neuer Eintrag erzeugt.
;================================================================
;	Definitionen

	processor	16c84
	radix		dec
	include		"p16cxx.inc"

;================================================================
;	Konfiguration

	org	H'2007'
	dw	_CP_OFF & _PWRTE_ON & _WDT_OFF & _LP_OSC

;================================================================
;	Konstanten

;ACHTUNG, Quarz mu 32768 Hz sein!

;================================================================
;	Ein- und Ausgnge

TRIS_A	equ	B'00000000'	;Port A:
LCD_A0	equ	0		;O LCD 0/1 = Befehl/Daten
LCD_CS	equ	1		;O LCD 0 = Freigabe
LCD_RD	equ	2		;O LCD 0 = Lesen
LCD_WR	equ	3		;O LCD 0 = Schreiben
;	equ	4		;O unbenutzt

TRIS_B	equ	B'00000111'	;Port B:
KEY_L	equ	0		;I 0 = linke Taste gedrckt
KEY_M	equ	1		;I 0 = mittlere Taste gedrckt
KEY_R	equ	2		;I 0 = rechte Taste gedrckt
;	equ	3		;O unbenutzt
;	equ	4..7		;O LCD Datenbus

KEY_ALL	equ	(1<<KEY_L)|(1<<KEY_M)|(1<<KEY_R)

OPT	equ	B'01010100'	;Pullups an Port B
				;TMR0 als Timer F_CYCLE/32

;================================================================
;	Daten
	org	H'0C'

flags:		res	1	;Diverse Flags (siehe fl_...)
fl_flash	equ	0	;fr blinkende Zeichen

prefix:		res	2	;Vorwahl

unit:		res	1	;Gebhr fr eine Minute
sum:		res	2	;Summe der Gebhren

line:		res	10	;Puffer fr ASCII-Ausgabe

cursor:		res	1	;Cursor beim Editieren

save_eeadr:	res	1	;Zwischenspeicher fr EEADR

count_timer:	res	1	;Teiler fr Minuten

hour:		res	1
min:		res	1	;Zeitzhler

lcd_p_i:	res	1
lcd_dat:	res	1	;LCD-Ausgabe

v_count:	res	1
v_bin:		res	2
v_bcd:		res	3	;Binr-Dezimal-Umrechnung

	if	$ > H'30'
	error	"Zu viel RAM benutzt"
	endif

;================================================================
;	EEPROM
	org	H'213F'

	dw	H'FF'		;Ende der Liste markieren

;================================================================
;	Vektoren
	org	H'0000'
	goto	reset

	dw	VERSION

	org	H'0004'		;geht ohne Sprung weiter...
;================================================================
;	Interrupts
	return			;unbenutzt
;================================================================
;	Initialisierung
reset:
	bsf	STATUS,RP0
	movlw	TRIS_A
	movwf	TRISA		;Port A programmieren
	movlw	TRIS_B
	movwf	TRISB		;Port B programmieren
	movlw	OPT
	movwf	OPTION_REG	;Optionen programmieren
	bcf	STATUS,RP0
;----------------------------------------------------------------
	call	lcd_init	;LCD initialisieren
;================================================================
;	Hauptfunktionen
select_first:
	clrf	EEADR		;Zeiger auf ersten Eintrag

select:
	call	read_entry
	call	entry_bindec
	call	show_line	;aktuellen Eintrag anzeigen

	call	key_none	;Tasten loslassen

select_loop:
	btfss	PORTB,KEY_L
	goto	select_next	;nchster Eintrag?
	btfss	PORTB,KEY_M
	goto	edit		;Eintrag bearbeiten?
	btfss	PORTB,KEY_R
	goto	count		;Gebhren zhlen?
	goto	select_loop

select_next:
	movlw	3
	addwf	EEADR
	call	read_entry	;nchsten Eintrag holen

	incf	prefix+0,w
	btfss	STATUS,Z
	goto	select		;Eintrag gefunden?

	goto	select_first	;am Ende zyklisch weiter
;----------------------------------------------------------------
edit:
	call	entry_bindec	;aktuellen Eintrag vorbereiten

	movlw	1
	movwf	cursor		;Stelle setzen

	bcf	PORTA,LCD_A0
	movlw	B'00001111'	;"cursor on"
	call	lcd_out
	bsf	PORTA,LCD_A0

edit_show:
	call	show_line
	movf	cursor,w
	call	lcd_goto	;Eintrag und Stelle anzeigen

	call	key_none	;Tasten loslassen

edit_loop:
	btfss	PORTB,KEY_L
	goto	edit_inc	;Digit zhlen?
	btfss	PORTB,KEY_M
	goto	special		;Spezialmen?
	btfss	PORTB,KEY_R
	goto	edit_next	;nchste Stelle?
	goto	edit_loop
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
edit_inc:
	movf	cursor,w
	addlw	line
	movwf	FSR
	incf	INDF		;Stelle erhhen

	movf	cursor,w
	xorlw	6
	movlw	'9'+1
	btfsc	STATUS,Z
	movlw	'1'+1		;Maximum bestimmen

	xorwf	INDF,w
	movlw	'0'
	btfsc	STATUS,Z
	movwf	INDF		;zyklisch wieder auf '0'

	goto	edit_show
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
edit_next:
	incf	cursor
	movf	cursor,w
	xorlw	5
	btfsc	STATUS,Z
	incf	cursor
	xorlw	5^7
	btfsc	STATUS,Z
	incf	cursor
	xorlw	7^10
	btfss	STATUS,Z
	goto	edit_show	;Cursor weitersetzen

	call	entry_decbin	;Eingabe zurckcodieren

	call	write_entry	;Eintrag speichern

	bcf	PORTA,LCD_A0
	movlw	B'00001110'	;"cursor off"
	call	lcd_out
	bsf	PORTA,LCD_A0

	goto	select
;----------------------------------------------------------------
special:
	bcf	PORTA,LCD_A0
	movlw	B'00001110'	;"cursor off"
	call	lcd_out
	bsf	PORTA,LCD_A0

	movlw	0
	call	lcd_goto	;Textausgabe vorbereiten

	movf	EEADR,w
	movwf	save_eeadr	;Zeiger auf Eintrag retten
sp_last_loop:
	movlw	3
	addwf	EEADR
	call	read_entry	;nchsten Eintrag holen

	incf	prefix+0,w
	btfss	STATUS,Z
	goto	sp_last_loop	;noch nicht Ende der Liste?

	movf	EEADR,w
	xorlw	63
	btfsc	STATUS,Z
	goto	special_nonew	;kein Platz mehr?

	xorlw	63^3
	btfsc	STATUS,Z
	goto	special_nodel	;nur noch erster Eintrag?

	movf	save_eeadr,w
	movwf	EEADR		;Zeiger zurcksetzen

	movlw	'L'
	call	lcd_out
	movlw	H'A9'
	call	lcd_out
	movlw	' '
	call	lcd_out
	movlw	'A'
	call	lcd_out
	movlw	'B'
	call	lcd_out
	movlw	'R'
	call	lcd_out
	movlw	' '
	call	lcd_out
	movlw	'N'
	call	lcd_out
	movlw	'E'
	call	lcd_out
	movlw	'U'
	call	lcd_out		;Men anzeigen

	call	key_none	;Tasten loslassen

sp_all_loop:
	btfss	PORTB,KEY_L
	goto	delete_entry	;Eintrag lschen?
	btfss	PORTB,KEY_M
	goto	select		;Editieren abbrechen?
	btfss	PORTB,KEY_R
	goto	new_entry	;neuer Eintrag?
	goto	sp_all_loop
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
special_nonew:
	movf	save_eeadr,w
	movwf	EEADR		;Zeiger zurcksetzen

	movlw	'L'
	call	lcd_out
	movlw	H'A9'
	call	lcd_out
	movlw	' '
	call	lcd_out
	movlw	'A'
	call	lcd_out
	movlw	'B'
	call	lcd_out
	movlw	'R'
	call	lcd_out
	movlw	' '
	call	lcd_out
	movlw	' '
	call	lcd_out
	movlw	' '
	call	lcd_out
	movlw	' '
	call	lcd_out		;Men anzeigen

	call	key_none	;Tasten loslassen

sp_nonew_loop:
	btfss	PORTB,KEY_L
	goto	delete_entry	;Eintrag lschen?
	btfss	PORTB,KEY_M
	goto	select		;Editieren abbrechen?
	goto	sp_nonew_loop
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
special_nodel:
	movf	save_eeadr,w
	movwf	EEADR		;Zeiger zurcksetzen

	movlw	' '
	call	lcd_out
	movlw	' '
	call	lcd_out
	movlw	' '
	call	lcd_out
	movlw	'A'
	call	lcd_out
	movlw	'B'
	call	lcd_out
	movlw	'R'
	call	lcd_out
	movlw	' '
	call	lcd_out
	movlw	'N'
	call	lcd_out
	movlw	'E'
	call	lcd_out
	movlw	'U'
	call	lcd_out		;Men anzeigen

	call	key_none	;Tasten loslassen

sp_nodel_loop:
	btfss	PORTB,KEY_M
	goto	select		;Editieren abbrechen?
	btfss	PORTB,KEY_R
	goto	new_entry	;neuer Eintrag?
	goto	sp_nodel_loop
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
delete_entry:
de_loop:
	movlw	3
	addwf	EEADR
	call	read_entry	;nchsten Eintrag holen

	movlw	3
	subwf	EEADR
	call	write_entry
	movlw	3
	addwf	EEADR		;ber vorherigen Eintrag schreiben

	incf	prefix+0,w
	btfss	STATUS,Z
	goto	de_loop		;noch nicht Ende der Liste?

	goto	select_first
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
new_entry:
ne_last_loop:
	movlw	3
	addwf	EEADR
	call	read_entry	;nchsten Eintrag holen

	incf	prefix+0,w
	btfss	STATUS,Z
	goto	ne_last_loop	;noch nicht Ende der Liste?

	movlw	1000 >> 8
	movwf	prefix+0
	movlw	1000 & H'FF'
	movwf	prefix+1
	movlw	1
	movwf	unit		;Werte vorbesetzen

	call	write_entry	;Eintrag speichern

	goto	edit
;----------------------------------------------------------------
count:
	clrf	sum+0
	clrf	sum+1
	clrf	hour
	clrf	min		;Werte lschen
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
count_run:
	call	key_none	;Tasten loslassen

	clrf	TMR0
	bcf	INTCON,T0IF	;Zeitnahme starten
count_next:
	movlw	60
	movwf	count_timer	;Minute zurcksetzen

	movf	unit,w
	addwf	sum+1
	btfsc	STATUS,C
	incf	sum+0		;Gebhrensumme weiterzhlen

	incf	min		;Minuten weiterzhlen
	movf	min,w
	xorlw	60
	btfss	STATUS,Z
	goto	count_show	;Stunde nicht voll?

	clrf	min
	incf	hour		;Stunde weiterzhlen
count_show:
	bsf	flags,fl_flash	;Trenner ist ein

	movlw	0
	call	lcd_goto

	clrf	v_bin+0
	movf	hour,w
	movwf	v_bin+1
	call	bindec
	movf	v_bcd+2,w
	andlw	H'0F'
	addlw	'0'
	call	lcd_out		;Stunden ausgeben

	movlw	':'
	call	lcd_out

	clrf	v_bin+0
	movf	min,w
	movwf	v_bin+1
	call	bindec
	swapf	v_bcd+2,w
	andlw	H'0F'
	addlw	'0'
	call	lcd_out
	movf	v_bcd+2,w
	andlw	H'0F'
	addlw	'0'
	call	lcd_out		;Minuten ausgeben

	movlw	' '
	call	lcd_out

	movf	sum+0,w
	movwf	v_bin+0
	movf	sum+1,w
	movwf	v_bin+1
	call	bindec
	swapf	v_bcd+1,w
	andlw	H'0F'
	addlw	'0'
	call	lcd_out
	movf	v_bcd+1,w
	andlw	H'0F'
	addlw	'0'
	call	lcd_out		;volle Whrung ausgeben

	movlw	'.'
	call	lcd_out

	swapf	v_bcd+2,w
	andlw	H'0F'
	addlw	'0'
	call	lcd_out
	movf	v_bcd+2,w
	andlw	H'0F'
	addlw	'0'
	call	lcd_out		;1/100 Whrung ausgeben

count_run_loop:
	btfss	PORTB,KEY_L
	goto	select		;neue Auswahl?
	btfss	PORTB,KEY_R
	goto	count_stop	;Stop?

	btfss	INTCON,T0IF
	goto	count_run_loop	;noch keine Sekunde?

	bcf	INTCON,T0IF

	movlw	1<<fl_flash
	xorwf	flags
	movlw	1
	call	lcd_goto
	movlw	' '
	btfsc	flags,fl_flash
	movlw	':'
	call	lcd_out		;Uhrzeit-Trenner blinkt

	decfsz	count_timer
	goto	count_run_loop	;noch keine Minute?

	goto	count_next
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
count_stop:
	movlw	1
	call	lcd_goto
	movlw	':'
	call	lcd_out		;Uhrzeit-Trenner an

	call	key_none	;Tasten loslassen

count_stop_loop:
	btfss	PORTB,KEY_L
	goto	select		;neue Auswahl?
	btfss	PORTB,KEY_R
	goto	count_run	;Start?

	goto	count_stop_loop
;================================================================
;	Eintrag holen
read_entry:
	bsf	STATUS,RP0
	bsf	EECON1,RD
	bcf	STATUS,RP0
	movf	EEDATA,w
	movwf	prefix+0
	incf	EEADR
	bsf	STATUS,RP0
	bsf	EECON1,RD
	bcf	STATUS,RP0
	movf	EEDATA,w
	movwf	prefix+1	;Vorwahl holen

	incf	EEADR
	bsf	STATUS,RP0
	bsf	EECON1,RD
	bcf	STATUS,RP0
	movf	EEDATA,w
	movwf	unit		;Gebhr holen

	movlw	2
	subwf	EEADR		;Zeiger zurcksetzen
	return
;================================================================
;	Eintrag speichern
write_entry:
	movf	prefix+0,w
	movwf	EEDATA
	bsf	STATUS,RP0
	bsf	EECON1,WREN
	bcf	EECON1,EEIF
	movlw	H'55'
	movwf	EECON2
	movlw	H'AA'
	movwf	EECON2
	bsf	EECON1,WR
we_loop1:
	btfss	EECON1,EEIF
	goto	we_loop1
	bcf	STATUS,RP0
	incf	EEADR
	movf	prefix+1,w
	movwf	EEDATA
	bsf	STATUS,RP0
	bcf	EECON1,EEIF
	movlw	H'55'
	movwf	EECON2
	movlw	H'AA'
	movwf	EECON2
	bsf	EECON1,WR
we_loop2:
	btfss	EECON1,EEIF
	goto	we_loop2
	bcf	STATUS,RP0	;Vorwahl speichern

	incf	EEADR
	movf	unit,w
	movwf	EEDATA
	bsf	STATUS,RP0
	bcf	EECON1,EEIF
	movlw	H'55'
	movwf	EECON2
	movlw	H'AA'
	movwf	EECON2
	bsf	EECON1,WR
we_loop3:
	btfss	EECON1,EEIF
	goto	we_loop3
	bcf	EECON1,WREN
	bcf	STATUS,RP0	;Gebhr speichern

	movlw	2
	subwf	EEADR		;Zeiger zurcksetzen
	return
;================================================================
;	Eintrag (binr) in Zeile (ASCII) umsetzen
entry_bindec:
	movf	prefix+0,w
	movwf	v_bin+0
	movf	prefix+1,w
	movwf	v_bin+1
	call	bindec

	movf	v_bcd+0,w
	andlw	H'0F'
	addlw	'0'
	movwf	line+0
	swapf	v_bcd+1,w
	andlw	H'0F'
	addlw	'0'
	movwf	line+1
	movf	v_bcd+1,w
	andlw	H'0F'
	addlw	'0'
	movwf	line+2
	swapf	v_bcd+2,w
	andlw	H'0F'
	addlw	'0'
	movwf	line+3
	movf	v_bcd+2,w
	andlw	H'0F'
	addlw	'0'
	movwf	line+4		;Vorwahl umsetzen

	movlw	' '
	movwf	line+5

	clrf	v_bin+0
	movf	unit,w
	movwf	v_bin+1
	call	bindec

	movf	v_bcd+1,w
	andlw	H'0F'
	addlw	'0'
	movwf	line+6
	movlw	'.'
	movwf	line+7
	swapf	v_bcd+2,w
	andlw	H'0F'
	addlw	'0'
	movwf	line+8
	movf	v_bcd+2,w
	andlw	H'0F'
	addlw	'0'
	movwf	line+9		;Gebhr umsetzen
	return
;================================================================
;	Zeile (ASCII) in Eintrag (binr) umsetzen
entry_decbin:
	movf	line+0,w
	andlw	H'0F'
	movwf	v_bcd+0
	swapf	line+1,w
	andlw	H'F0'
	movwf	v_bcd+1
	movf	line+2,w
	andlw	H'0F'
	iorwf	v_bcd+1
	swapf	line+3,w
	andlw	H'F0'
	movwf	v_bcd+2
	movf	line+4,w
	andlw	H'0F'
	iorwf	v_bcd+2

	call	decbin
	movf	v_bin+0,w
	movwf	prefix+0
	movf	v_bin+1,w
	movwf	prefix+1	;Vorwahl umsetzen

	clrf	v_bcd+0
	movf	line+6,w
	andlw	H'0F'
	movwf	v_bcd+1
	swapf	line+8,w
	andlw	H'F0'
	movwf	v_bcd+2
	movf	line+9,w
	andlw	H'0F'
	iorwf	v_bcd+2

	call	decbin
	movf	v_bin+1,w
	movwf	unit		;Gebhr umsetzen
	return
;================================================================
;	Binrzahl in BCD umrechnen
bindec:
	clrf	v_bcd+0
	clrf	v_bcd+1
	clrf	v_bcd+2
	movlw	16
	movwf	v_count		;Umrechnung vorbereiten
bd_loop:
	movlw	H'03'
	addwf	v_bcd+0
	btfss	v_bcd+0,3
	subwf	v_bcd+0
	addwf	v_bcd+1
	btfss	v_bcd+1,3
	subwf	v_bcd+1
	addwf	v_bcd+2
	btfss	v_bcd+2,3
	subwf	v_bcd+2		;BCD-Korrektur unteres Nibble

	movlw	H'30'
	addwf	v_bcd+1
	btfss	v_bcd+1,7
	subwf	v_bcd+1
	addwf	v_bcd+2
	btfss	v_bcd+2,7
	subwf	v_bcd+2		;BCD-Korrektur oberes Nibble

	rlf	v_bin+1
	rlf	v_bin+0
	rlf	v_bcd+2
	rlf	v_bcd+1
	rlf	v_bcd+0		;Bit bergeben

	decfsz	v_count
	goto	bd_loop		;noch nicht alle Bits?
	return
;================================================================
;	BCD-Zahl in binr umrechnen
decbin:
	movlw	16
	movwf	v_count		;Umrechnung vorbereiten
db_loop:
	bcf	STATUS,C
	rrf	v_bcd+0
	rrf	v_bcd+1
	rrf	v_bcd+2
	rrf	v_bin+0
	rrf	v_bin+1		;Bit bergeben

	movlw	H'03'
	btfsc	v_bcd+0,3
	subwf	v_bcd+0
	btfsc	v_bcd+1,3
	subwf	v_bcd+1
	btfsc	v_bcd+2,3
	subwf	v_bcd+2		;BCD-Korrektur unteres Nibble

	movlw	H'30'
	btfsc	v_bcd+1,7
	subwf	v_bcd+1
	btfsc	v_bcd+2,7
	subwf	v_bcd+2		;BCD-Korrektur oberes Nibble

	decfsz	v_count
	goto	db_loop		;noch nicht alle Bits?
	return
;================================================================
;	ASCII-Puffer ausgeben
show_line:
	movlw	0
	call	lcd_goto

	movf	line+0,w
	call	lcd_out
	movf	line+1,w
	call	lcd_out
	movf	line+2,w
	call	lcd_out
	movf	line+3,w
	call	lcd_out
	movf	line+4,w
	call	lcd_out
	movf	line+5,w
	call	lcd_out
	movf	line+6,w
	call	lcd_out
	movf	line+7,w
	call	lcd_out
	movf	line+8,w
	call	lcd_out
	movf	line+9,w
	call	lcd_out
	return
;================================================================
;	LCD initialisieren
lcd_init:
	bsf	PORTA,LCD_CS	;Peripherie vorbereiten

	bcf	PORTA,LCD_A0
	movlw	B'00010000'	;"reset"
	call	lcd_out
	movlw	B'00010010'	;"1 line"
	call	lcd_out
	movlw	B'00001101'	;"display on"
	call	lcd_out
	movlw	B'00001011'	;"cursor blinking"
	call	lcd_out
	movlw	B'00001001'	;"cursor block"
	call	lcd_out
	bsf	PORTA,LCD_A0	;LCD initialisieren
	return
;================================================================
;	Cursor setzen
lcd_goto:
	bcf	PORTA,LCD_A0
	addlw	B'10000000'	;"1st line"
	call	lcd_out
	bsf	PORTA,LCD_A0	;Cursor setzen
	return
;================================================================
;	LCD schreiben
lcd_out:
	movwf	lcd_dat		;Wert retten

	bcf	PORTA,LCD_CS
	movwf	PORTB
	bcf	PORTA,LCD_WR
	bsf	PORTA,LCD_WR	;oberes Nibble ausgeben

	swapf	lcd_dat,w
	movwf	PORTB
	bcf	PORTA,LCD_WR
	bsf	PORTA,LCD_WR	;unteres Nibble ausgeben
	bsf	PORTA,LCD_CS
	return
;================================================================
;	Loslassen aller Tasten abwarten
key_none:
	movf	PORTB,w
	andlw	KEY_ALL		;alle Tastenbits holen

	xorlw	KEY_ALL
	btfss	STATUS,Z
	goto	key_none	;noch eine Taste gedrckt?

	return
;================================================================
	end
